/*
 * upf_usbgem.c : ADMtek an986/adm8511/8515 USB to Fast Ethernet Driver
 * for Solaris
 */

/*
 * Copyright (c) 2004-2008 Masayuki Murayama.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the author nor the names of its contributors may be
 *    used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

#pragma ident   "%Z%%M% %I%     %E%"

/*
 *  Changelog:
 */

/*
 * TODO
 */
/* ======================================================= */

/*
 * Solaris system header files and macros
 */
#include <sys/types.h>
#include <sys/conf.h>
#include <sys/debug.h>
#include <sys/kmem.h>
#include <sys/modctl.h>
#include <sys/errno.h>
#include <sys/ddi.h>
#include <sys/sunddi.h>
#include <sys/byteorder.h>
#include <sys/ethernet.h>
#include <sys/pci.h>
#include <sys/gld.h>
#include <sys/usb/usba.h>

#include "usbgem.h"
#include "mii.h"
#include "adm8511reg.h"

char	ident[] = "an986/adm8511/adm8515 usbnic driver v" VERSION;
char	_depends_on[] = {"misc/gld misc/usba"};

/*
 * Useful macros
 */
#ifndef B_FALSE
#define	B_FALSE	(0)
#endif
#ifndef B_TRUE
#define	B_TRUE	(!B_FALSE)
#endif
#define	ROUNDUP2(x, y)	(((x)+(y)-1) & ~((y)-1))
#define	CHECK_AND_JUMP(val, label)	\
		if ((val) != USB_SUCCESS) { goto label; }

#define	IS_ADM8513(lp)	((lp)->chip_type == CHIP_ADM8513)

/*
 * Debugging
 */
#ifdef DEBUG_LEVEL
static int upf_debug = DEBUG_LEVEL;
#define	DPRINTF(n, args)	if (upf_debug > (n)) cmn_err args
#else
#define	DPRINTF(n, args)
#endif

/*
 * Our configration for ADMtek Pegasus/PegasusII
 */
/* timeouts */
#define	ONESEC		(drv_usectohz(1*1000000))

/*
 * RX/TX buffer size
 */

#define	TXBUF_LEN	ROUNDUP2(ETHERMAX, 4)

/*
 * Local device definitions
 */
struct upf_dev {
	/*
	 * Misc HW information
	 */
	uint8_t			ectrl0;
	uint8_t			ectrl1;
	uint8_t			ectrl2;
	uint8_t			mac_addr[ETHERADDRL];
	int			chip_type;
#define	CHIP_ADM8511	0	/* including an986 and adm8515 */
#define	CHIP_ADM8513	1
	boolean_t		init_done;
	boolean_t		power_on_mac_valid;
};

/*
 * private functions
 */

/* mii operations */
static void  upf_mii_sync(struct usbgem_dev *, int *errp);
static uint16_t  upf_mii_read(struct usbgem_dev *, uint_t, int *errp);
static void upf_mii_write(struct usbgem_dev *, uint_t, uint16_t, int *errp);

/* nic operations */
static int upf_attach_chip(struct usbgem_dev *);
static int upf_reset_chip(struct usbgem_dev *);
static int upf_init_chip(struct usbgem_dev *);
static int upf_start_chip(struct usbgem_dev *);
static int upf_stop_chip(struct usbgem_dev *);
static int upf_set_media(struct usbgem_dev *);
static int upf_set_rx_filter(struct usbgem_dev *);
static int upf_get_stats(struct usbgem_dev *);

/* packet operations */
static mblk_t *upf_tx_make_packet(struct usbgem_dev *, mblk_t *);
static mblk_t *upf_rx_make_packet(struct usbgem_dev *, mblk_t *);

/* interrupt handler */

/* =============================================================== */
/*
 * I/O functions
 */
/* =============================================================== */
#define	UPF_REQ_GET_REGISTER	0xf0
#define	UPF_REQ_SET_REGISTER	0xf1
#define	OUTB(dp, p, v, errp, label)	\
	if ((*(errp) = usbgem_ctrl_out((dp), 	\
	/* bmRequestType */ USB_DEV_REQ_HOST_TO_DEV	\
		| USB_DEV_REQ_TYPE_VENDOR | USB_DEV_REQ_RCPT_DEV,	\
	/* bRequest */	UPF_REQ_SET_REGISTER,	\
	/* wValue */	(v),	\
	/* wIndex */	(p),	\
	/* wLength */	1,	\
	/* buf */	NULL,	\
	/* size */	0)) != USB_SUCCESS) goto label;

#define	OUTW(dp, p, v, errp, label)	\
	if ((*(errp) = usbgem_ctrl_out_val((dp), 	\
	/* bmRequestType */ USB_DEV_REQ_HOST_TO_DEV	\
		| USB_DEV_REQ_TYPE_VENDOR | USB_DEV_REQ_RCPT_DEV,	\
	/* bRequest */	UPF_REQ_SET_REGISTER,	\
	/* wValue */	0,	\
	/* wIndex */	(p),	\
	/* wLength */	2,	\
	/* value */	(v))) != USB_SUCCESS) goto label

#define	OUTS(dp, p, buf, len, errp, label)	\
	if ((*(errp) = usbgem_ctrl_out((dp), 	\
	/* bmRequestType */ USB_DEV_REQ_HOST_TO_DEV	\
		| USB_DEV_REQ_TYPE_VENDOR | USB_DEV_REQ_RCPT_DEV,	\
	/* bRequest */	UPF_REQ_SET_REGISTER,	\
	/* wValue */	0,	\
	/* wIndex */	(p),	\
	/* wLength */	(len),	\
	/* buf */	(buf),	\
	/* size */	(len))) != USB_SUCCESS) goto label

#define	INB(dp, p, vp, errp, label)	\
	if ((*(errp) = usbgem_ctrl_in_val((dp), 	\
	/* bmRequestType */ USB_DEV_REQ_DEV_TO_HOST	\
		| USB_DEV_REQ_TYPE_VENDOR | USB_DEV_REQ_RCPT_DEV,	\
	/* bRequest */ UPF_REQ_GET_REGISTER,	\
	/* wValue */	0,	\
	/* wIndex */	(p),	\
	/* wLength */	1,	\
	/* valuep */	(vp))) != USB_SUCCESS) goto label

#define	INW(dp, p, vp, errp, label)	\
	if ((*(errp) = usbgem_ctrl_in_val((dp), 	\
	/* bmRequestType */ USB_DEV_REQ_DEV_TO_HOST	\
		| USB_DEV_REQ_TYPE_VENDOR | USB_DEV_REQ_RCPT_DEV,	\
	/* bRequest */ UPF_REQ_GET_REGISTER,	\
	/* wValue */	0,	\
	/* wIndex */	(p),	\
	/* wLength */	2,	\
	/* valuep */	(vp))) != USB_SUCCESS) goto label

#define	INS(dp, p, buf, len, errp, label)	\
	if ((*(errp) = usbgem_ctrl_in((dp), 	\
	/* bmRequestType */ USB_DEV_REQ_DEV_TO_HOST	\
		    | USB_DEV_REQ_TYPE_VENDOR | USB_DEV_REQ_RCPT_DEV,	\
	/* bRequest */ UPF_REQ_GET_REGISTER,	\
	/* wValue */	0,	\
	/* wIndex */	(p),	\
	/* wLength */	(len),	\
	/* buf */	(buf),	\
	/* size */	(len))) != USB_SUCCESS) goto label

/* =============================================================== */
/*
 * variables
 */
/* =============================================================== */
static int upf_ppa = 0;

/* =============================================================== */
/*
 * Hardware manupilation
 */
/* =============================================================== */
static int
upf_reset_chip(struct usbgem_dev *dp)
{
	int		i;
	uint8_t		val;
	int		err;
	struct upf_dev	*lp = dp->private;

	DPRINTF(0, (CE_CONT, "!%s: %s: called", dp->name, __func__));
	bzero(lp->mac_addr, sizeof (lp->mac_addr));

	lp->ectrl1 = 0;
	OUTB(dp, ECTRL1, ECTRL1_RESET, &err, usberr);

	for (i = 0; i < 1000; i++) {
		INB(dp, ECTRL1, &val, &err, usberr);
		if ((val & ECTRL1_RESET) == 0) {
			lp->ectrl1 = val;
			goto reset_done;
		}
		drv_usecwait(10);
	}
	/* time out */
	cmn_err(CE_WARN, "!%s: failed to reset: timeout", dp->name);
	return (USB_FAILURE);

reset_done:
	/*
	 * enable internal phy
	 */
	INB(dp, IPHY, &val, &err, usberr);
	val = (val | IPHY_EN) & ~IPHY_RST;
	OUTB(dp, IPHY, val, &err, usberr);

	INB(dp, IPHY, &val, &err, usberr);
	DPRINTF(0, (CE_CONT, "!%s: %s: IPHY: %02x", dp->name, __func__, val));
	if (val != 0) {
		/* reset internal phy on power-up */
		if (!lp->init_done) {
			OUTB(dp, IPHY, val | IPHY_RST, &err, usberr);
			OUTB(dp, IPHY, val, &err, usberr);
		}
	} else {
		/*
		 * It should be AN986 which doesn't have an internal PHY.
		 * We need to setup gpio ports in AN986, which are
		 * connected to external PHY control pins.
		 */
		if (!lp->init_done) {
			/* reset external phy on power-up */
			/* output port#0 L, port#1 L */
			OUTB(dp, GPIO10, GPIO10_0O | GPIO10_0OE, &err, usberr);
		}
		/* output port#0 H, port#1 L */
		OUTB(dp, GPIO10,
		    GPIO10_0O | GPIO10_0OE | GPIO10_1OE, &err, usberr);
	}
	lp->init_done = B_TRUE;
#if 0
	drv_usecwait(100*1000);
#endif
	return (USB_SUCCESS);

usberr:
	cmn_err(CE_NOTE, "!%s: %s: usberr detected", dp->name, __func__);
	return (USB_FAILURE);
}

/*
 * Setup an986/adm8511/adm8513/adm8515
 */
static int
upf_init_chip(struct usbgem_dev *dp)
{
	int		i;
	uint32_t	val;
	int		err;
	struct upf_dev	*lp = dp->private;

	DPRINTF(0, (CE_CONT, "!%s: %s: called", dp->name, __func__));

	/* ethernet control register 0 */
	lp->ectrl0 = ECTRL0_RX_FLOW | ECTRL0_RXSTAT | ECTRL0_CRC;
	OUTB(dp, ECTRL0, lp->ectrl0, &err, usberr);

	/* ethernet control reg1: will be set later in set_rx_filter() */
	lp->ectrl1 = 0;		/* default is 10M half duplex */

	/* ethernet control register 2: will be set later in set_rx_filter() */
	lp->ectrl2 = ECTRL2_RXBAD;	/* */
	OUTB(dp, ECTRL2, lp->ectrl2, &err, usberr);

	/* Multicast address: will be set later in upf_set_rx_filter() */

	/* Ethernet ID : will be set later in upf_set_rx_filter() */

	/* PAUSE timer low: */
	OUTB(dp, PAUSE, 0x1f, &err, usberr);

	/* receive packet number based pause control */
	/* XX */

	/* occupied receive FIFO based pause control */
	/* XX */

	/* EP1 control */
	/* default */

	/* Rx FIFO control */
	/* default (internal 24k fifo) */

	/* BIST contror: do nothing */

	err = upf_set_media(dp);
	CHECK_AND_JUMP(err, usberr);

	/* Configuration register 0: no need to change */

	DPRINTF(2, (CE_CONT, "!%s: upf_start_chip: end (success)", dp->name));
	return (USB_SUCCESS);

usberr:
	cmn_err(CE_NOTE, "!%s: %s: usberr(%d) detected",
	    dp->name, __func__, err);
	return (USB_FAILURE);
}

static int
upf_start_chip(struct usbgem_dev *dp)
{
	int		err;
	struct upf_dev	*lp = dp->private;

	/* enable RX and TX */
	lp->ectrl0 |= ECTRL0_TX | ECTRL0_RX;
	OUTB(dp, ECTRL0, lp->ectrl0, &err, usberr);
	return (USB_SUCCESS);

usberr:
	cmn_err(CE_WARN, "!%s: %s: usberr(%d) detected",
	    dp->name, __func__, err);
	return (err);
}

static int
upf_stop_chip(struct usbgem_dev *dp)
{
	int		err;
	struct upf_dev	*lp = dp->private;

	/* disable RX and TX */
	lp->ectrl0 &= ~(ECTRL0_TX | ECTRL0_RX);
	OUTB(dp, ECTRL0, lp->ectrl0, &err, usberr);

	if (lp->power_on_mac_valid) {
		/* restore factory address */
		OUTS(dp, ETHERNETID,
		    &dp->dev_addr.ether_addr_octet[0],
		    ETHERADDRL, &err, usberr);
	} else {
		/* clear mac address */
		OUTS(dp, ETHERNETID,
		    "\000\000\000\000\000\000",
		    ETHERADDRL, &err, usberr);
	}

	return (USB_SUCCESS);

usberr:
	cmn_err(CE_WARN, "!%s: %s: usberr(%d) detected",
	    dp->name, __func__, err);
	return (err);
}

static int
upf_get_stats(struct usbgem_dev *dp)
{
	/* do nothing */
	return (USB_SUCCESS);
}

static uint_t
upf_mcast_hash(struct usbgem_dev *dp, uint8_t *addr)
{
	return (usbgem_ether_crc_le(addr));
}

static int
upf_set_rx_filter(struct usbgem_dev *dp)
{
	uint8_t		mhash[8];
	int		i;
	uint_t		h;
	int		err;
#ifdef DEBUG_LEVEL
	uint8_t		reg0;
	uint8_t		reg1;
	uint8_t		reg2;
#endif
	struct upf_dev	*lp = dp->private;

	DPRINTF(2, (CE_CONT, "!%s: %s: called, rxmode:%x",
	    dp->name, __func__, dp->rxmode));

	/* clear hash table and reset rx mode */
	bzero(mhash, sizeof (mhash));
	lp->ectrl0 &= ~ECTRL0_ALLMULTI;
	lp->ectrl2 &= ~ECTRL2_PROMISC;

	if ((dp->rxmode & RXMODE_PROMISC) != 0) {
		/* promiscious mode implies all multicast and all physical */
		lp->ectrl0 |= ECTRL0_ALLMULTI;
		lp->ectrl2 |= ECTRL2_PROMISC;
	} else if ((dp->rxmode & RXMODE_ALLMULTI) != 0 ||
	    dp->mc_count > 64/2) {
		/* accept all multicast packets */
		lp->ectrl0 |= ECTRL0_ALLMULTI;
	} else if (dp->mc_count > 0) {
		/*
		 * make hash table to receive interresting
		 * multicast addresses only.
		 */
		for (i = 0; i < dp->mc_count; i++) {
			/* hash table is 64 = 2^6 bit width */
			h = dp->mc_list[i].hash >> (32 - 6);
			mhash[h / 8] |= 1 << (h % 8);
		}
	}

	/* disable rx filter before changing it. */
	if ((lp->ectrl0 & ECTRL0_RX) != 0) {
		OUTB(dp, ECTRL0, lp->ectrl0 & ~ECTRL0_RX, &err, usberr);
	}

	/* update mac address */
	OUTS(dp, ETHERNETID,
	    &dp->cur_addr.ether_addr_octet[0], ETHERADDRL, &err, usberr);

	/* update multicast hash table */
	OUTS(dp, MCAST, mhash, 8, &err, usberr);

	/* update rx mode */
	OUTB(dp, ECTRL2, lp->ectrl2, &err, usberr);

	/* enable rx again */
	if ((lp->ectrl0 & ECTRL0_RX) != 0) {
		OUTB(dp, ECTRL0, lp->ectrl0, &err, usberr);
	}

#if DEBUG_LEVEL > 2
	INB(dp, ECTRL0, &reg0, &err, usberr);
	INB(dp, ECTRL1, &reg1, &err, usberr);
	INB(dp, ECTRL2, &reg2, &err, usberr);

	cmn_err(CE_CONT, "!%s: %s: returned, ectrl0:%b %b %b",
	    dp->name, __func__,
	    reg0, ECTRL0_BITS, reg1, ECTRL1_BITS, reg2, ECTRL2_BITS);
#endif
	return (USB_SUCCESS);

usberr:
	cmn_err(CE_NOTE, "!%s: %s: usberr detected", dp->name, __func__);

	/* try to enable RX */
	if ((lp->ectrl0 & ECTRL0_RX) != 0) {
		OUTB(dp, ECTRL0, lp->ectrl0, &err, usberr1);
	}
usberr1:
	return (USB_SUCCESS);
}

static int
upf_set_media(struct usbgem_dev *dp)
{
	uint8_t		new;
	int		err;
	struct upf_dev	*lp = dp->private;

	DPRINTF(2, (CE_CONT, "!%s: %s: called", dp->name, __func__));

	new = lp->ectrl1 & ~(ECTRL1_FDPX | ECTRL1_100M);

	/* select duplex */
	if (dp->full_duplex) {
		new |= ECTRL1_FDPX;
	}

	/* select speed */
	if (dp->speed == USBGEM_SPD_100) {
		new |= ECTRL1_100M;
	}

	/* flow control: rx flow contrl is always enabled. */

	/* update ether control1 register */
	if (new != lp->ectrl1) {
		lp->ectrl1 = new;
		OUTB(dp, ECTRL1, lp->ectrl1, &err, usberr);
	}

#ifdef notyet
	/* special treatment for Linksys products */
	if ((lp->flags & LINKSYS) != 0) {
		new = upf_mii_read(dp, 0x1b, &err) | 0x4;
		upf_mii_write(dp, 0x1b, new, &err);
	}
#endif

	DPRINTF(2, (CE_CONT, "!%s: %s: returned, ectrl1:%b",
	    dp->name, __func__, lp->ectrl1, ECTRL1_BITS));

	return (USB_SUCCESS);

usberr:
	cmn_err(CE_WARN, "%s: %s: failed to write ectrl1", dp->name, __func__);
	return (err);
}

/*
 * send/receive packet check
 */
static mblk_t *
upf_tx_make_packet(struct usbgem_dev *dp, mblk_t *mp)
{
	size_t		len;
	mblk_t		*new;
	mblk_t		*tp;
	uint8_t		*bp;
	int		msglen;

	len = msgdsize(mp);

	if (len > dp->mtu + sizeof (struct ether_header) + 4) {
		/* packet is too long */
		return (NULL);
	}

	msglen = len + sizeof (uint16_t);

	/* avoid usb controller bug */
	if ((msglen & 0x3f) == 0) {
		msglen++;
	}

	if ((new = allocb(msglen, 0)) == NULL) {
		return (NULL);
	}

	/* copy contents of the buffer */
	new->b_wptr = new->b_rptr + msglen;
	bp = new->b_rptr;

	/* the nic requires a two byte header of the packet size */
	bp[0] = (uint8_t)len;
	bp[1] = (uint8_t)(len >> 8);
	bp += sizeof (uint16_t);

	/* copy the payload */
	for (tp = mp; tp; tp = tp->b_cont) {
		len = tp->b_wptr - tp->b_rptr;
		if (len > 0) {
			bcopy(tp->b_rptr, bp, len);
			bp += len;
		}
	}

	return (new);
}

static void
upf_dump_packet(struct usbgem_dev *dp, uint8_t *bp, int n)
{
	int	i;

	for (i = 0; i < n; i += 8, bp += 8) {
		cmn_err(CE_CONT, "%02x %02x %02x %02x %02x %02x %02x %02x",
		    bp[0], bp[1], bp[2], bp[3], bp[4], bp[5], bp[6], bp[7]);
	}
}

static mblk_t *
upf_rx_make_packet(struct usbgem_dev *dp, mblk_t *mp)
{
	uint8_t		*p;
	uint16_t	rxhd;
	uint_t		len;
	uint8_t		rsr;
	struct upf_dev	*lp = dp->private;

	ASSERT(mp != NULL);

#ifdef DEBUG_LEVEL
	len = msgdsize(mp);
	DPRINTF(2, (CE_CONT, "!%s: time:%d %s: cont:%p",
	    dp->name, ddi_get_lbolt(), __func__, len, mp->b_cont));

	if (upf_debug > 3) {
		upf_dump_packet(dp, mp->b_rptr, max(6, len));
	}
#endif
	/* get the length of Rx packet */
	p = mp->b_wptr - 4;
	rsr = p[3];
	if (IS_ADM8513(lp)) {
		/* As Rx packets from ADM8153 have two byte header, remove it */
		p = mp->b_rptr;
		len = ((p[1] << 8) | p[0]) & 0x0fff;
		mp->b_rptr += 2;
	} else {
		len = (((p[1] << 8) | p[0]) & 0x0fff)- ETHERFCSL - 4;
	}

	DPRINTF(2, (CE_CONT, "!%s: %s: rsr:%b len:%d",
	    dp->name, __func__, rsr, RSR_BITS, len));

	/* check if error happen */
	if ((rsr & RSR_ERRORS) != 0) {
		DPRINTF(0, (CE_CONT, "!%s: rsr:%b", dp->name, rsr, RSR_BITS));
		if ((rsr & (RSR_CRC | RSR_DRIBBLE)) != 0) {
			dp->stats.frame++;
		}
		if ((rsr & RSR_LONG) != 0) {
			dp->stats.frame_too_long++;
		}
		if ((rsr & RSR_RUNT) != 0) {
			dp->stats.runt++;
		}

		dp->stats.errrcv++;
		return (NULL);
	}

	/* check packet size */
	if (len > ETHERMAX) {
		/* too long */
		dp->stats.frame_too_long++;
		dp->stats.errrcv++;
		return (NULL);
	} else if (len < ETHERMIN) {
		dp->stats.runt++;
		dp->stats.errrcv++;
		return (NULL);
	}

	/* remove tailing crc and rx status fields */
	mp->b_wptr = mp->b_rptr + len;
	return (mp);
}

/*
 * Device depend interrupt handler
 */
/* none */

/*
 * MII Interfaces
 */
static void
upf_mii_sync(struct usbgem_dev *dp, int *errp)
{
	DPRINTF(4, (CE_CONT, "!%s: %s: called", dp->name, __func__));

	/* do nothing */
	*errp = USB_SUCCESS;
}

static uint16_t
upf_mii_read(struct usbgem_dev *dp, uint_t index, int *errp)
{
	uint8_t		phyctrl;
	uint16_t	val;
	int		i;

	DPRINTF(4, (CE_CONT, "!%s: %s: called, ix:%d",
	    dp->name, __func__, index));
	ASSERT(index >= 0 && index < 32);

	*errp = USB_SUCCESS;

	/* set PHYADDR */
	OUTB(dp, PHYADDR, dp->mii_phy_addr, errp, usberr);

	/* Initiate MII read transaction */
	OUTB(dp, PHYCTRL, (index | PHYCTRL_RD), errp, usberr);

	for (i = 0; i < 100; i++) {
		INB(dp, PHYCTRL, &phyctrl, errp, usberr);
		if ((phyctrl & PHYCTRL_DONE) != 0) {
			/* done */
			INW(dp, PHYDATA, &val, errp, usberr);
			DPRINTF(4, (CE_CONT, "!%s: %s: return %04x",
			    dp->name, __func__, val));
			return (val);
		}
		drv_usecwait(10);
	}
	/* timeout */
	cmn_err(CE_WARN, "!%s: %s: timeout detected", dp->name, __func__);
	*errp = USB_FAILURE;
	return (0);

usberr:
	cmn_err(CE_CONT,
	    "!%s: %s: usberr(%d) detected", dp->name, __func__, *errp);
	return (0);
}

static void
upf_mii_write(struct usbgem_dev *dp, uint_t index, uint16_t val, int *errp)
{
	int		i;
	uint8_t		phyctrl;

	DPRINTF(5, (CE_CONT, "!%s: %s called", dp->name, __func__));
	ASSERT(index >= 0 && index < 32);

	*errp = USB_SUCCESS;

	OUTW(dp, PHYDATA, val, errp, usberr);
	OUTB(dp, PHYADDR, dp->mii_phy_addr, errp, usberr);
	OUTB(dp, PHYCTRL, index | PHYCTRL_WR, errp, usberr);

	for (i = 0; i < 100; i++) {
		INB(dp, PHYCTRL, &phyctrl, errp, usberr);
		if ((phyctrl & PHYCTRL_DONE) != 0) {
			/* done */
			return;
		}
		drv_usecwait(10);
	}

	/* time out */
	cmn_err(CE_WARN, "!%s: %s: timeout detected", dp->name, __func__);
	*errp = USB_FAILURE;
	return;

usberr:
	cmn_err(CE_CONT,
	    "!%s: %s: usberr(%d) detected", dp->name, __func__, *errp);
}

/* ======================================================== */
/*
 * OS depend (device driver DKI) routine
 */
/* ======================================================== */
static uint16_t
upf_read_eeprom(struct usbgem_dev *dp, int index, int *errp)
{
	int		i;
	uint8_t		eectrl;
	uint16_t	data;

	*errp = USB_SUCCESS;

	OUTB(dp, EECTRL, 0, errp, usberr);

	OUTB(dp, EEOFFSET, index, errp, usberr);
	OUTB(dp, EECTRL, EECTRL_RD, errp, usberr);

	for (i = 0; i < 100; i++) {
		INB(dp, EECTRL, &eectrl, errp, usberr);
		if ((eectrl & EECTRL_DONE) != 0) {
			INW(dp, EEDATA, &data, errp, usberr);
			return (data);
		}
		drv_usecwait(10);
	}

	/* time out */
	*errp = USB_FAILURE;
	return (0);

usberr:
	cmn_err(CE_CONT,
	    "!%s: %s: usberr(%d) detected", dp->name, __func__, *errp);
	return (0);
}

static void
upf_eeprom_dump(struct usbgem_dev *dp, int size)
{
	int	i;
	int	err;

	cmn_err(CE_CONT, "!%s: %s dump:", dp->name, __func__);

	for (i = 0; i < size; i += 4) {
		cmn_err(CE_CONT, "!0x%02x: 0x%04x 0x%04x 0x%04x 0x%04x",
		    i*2,
		    upf_read_eeprom(dp, i + 0, &err),
		    upf_read_eeprom(dp, i + 1, &err),
		    upf_read_eeprom(dp, i + 2, &err),
		    upf_read_eeprom(dp, i + 3, &err));
	}
}

static int
upf_attach_chip(struct usbgem_dev *dp)
{
	int		i;
	int		err;
	uint16_t	val;
	uint8_t		*mac;
	uint8_t		ff6[] = {0xff, 0xff, 0xff, 0xff, 0xff, 0xff};
	uint8_t		zero6[] = {0, 0, 0, 0, 0, 0};
	struct upf_dev	*lp = dp->private;

	/*
	 * check chip type
	 */
	OUTB(dp, 0x83, 0xa5, &err, usberr);
	INB(dp, 0x83, &val, &err, usberr);

	DPRINTF(-1, (CE_CONT, "%s: %s: reg83:%x", dp->name, __func__, val));

	if (val == 0xa5) {
		lp->chip_type = CHIP_ADM8513;
		cmn_err(CE_CONT,
		    "!%s: %s: the device is ADM8513", dp->name, __func__);
	} else {
		lp->chip_type = CHIP_ADM8511;
		cmn_err(CE_CONT,
		    "!%s: %s: the device is %s",
		    dp->name, __func__, "AN986/ADM8511/ADM8515");
	}

	/* test if mac address is loaded from eeprom */
	mac = &dp->dev_addr.ether_addr_octet[0];
	INS(dp, ETHERNETID, mac, ETHERADDRL, &err, usberr);
	lp->power_on_mac_valid = (mac[0] & 3) == 0 &&
	    bcmp(mac, ff6, ETHERADDRL) != 0 &&
	    bcmp(mac, zero6, ETHERADDRL) != 0;

	DPRINTF(0, (CE_CONT,
	    "%s: %s: mac(power on): %02x:%02x:%02x:%02x:%02x:%02x",
	    dp->name, __func__,
	    mac[0], mac[1], mac[2], mac[3], mac[4], mac[5]));

	if (!lp->power_on_mac_valid) {
		/*
		 * The device is first generation of pegasus, not pegasus II.
		 * Read mac address from EEPROM
		 */
		DPRINTF(0, (CE_CONT,
		    "!%s: %s: reading mac address from EEPROM",
		    dp->name, __func__));
		for (i = 0; i < 3; i++) {
			val = upf_read_eeprom(dp, i, &err);
			if (err != USB_SUCCESS) {
				goto usberr;
			}
			mac[i*2+0] = (uint8_t)val;
			mac[i*2+1] = (uint8_t)(val >> 8);
		}

		DPRINTF(0, (CE_CONT,
		    "%s: %s: mac: %02x:%02x:%02x:%02x:%02x:%02x",
		    dp->name, __func__,
		    mac[0], mac[1], mac[2], mac[3], mac[4], mac[5]));
	}

	/* we dont use interrupt messages */
	dp->misc_flag |= USBGEM_NOINTR;

#if DEBUG_LEVEL > 3
	upf_eeprom_dump(dp, 0x80);
#endif
	return (USB_SUCCESS);

usberr:
	cmn_err(CE_WARN, "!%s: %s: usb error detected", dp->name, __func__);
	return (USB_FAILURE);
}

static int
upfattach(dev_info_t *dip, ddi_attach_cmd_t cmd)
{
	int			i;
	ddi_iblock_cookie_t	c;
	int			ret;
	int			vid;
	int			did;
	int			revid;
	int			unit;
	uint32_t		tcr;
	int			len;
	const char		*drv_name;
	struct usbgem_dev	*dp;
	void			*base;
	struct usbgem_conf	*ugcp;
	struct upf_dev		*lp;

	unit =  ddi_get_instance(dip);
	drv_name = ddi_driver_name(dip);

	DPRINTF(3, (CE_CONT, "!%s%d: %s: called, cmd:%d",
	    drv_name, unit, __func__, cmd));

	if (cmd == DDI_ATTACH) {
		/*
		 * Check if the chip is supported.
		 */

		/*
		 * Check the chip if it is really realtek rtl8150
		 */

		/*
		 * construct usbgem configration
		 */
		ugcp = kmem_zalloc(sizeof (*ugcp), KM_SLEEP);

		/* name */
		sprintf(ugcp->usbgc_name, "%s%d", drv_name, upf_ppa);
		ugcp->usbgc_ppa = upf_ppa;

		ugcp->usbgc_ifnum = 0;
		ugcp->usbgc_alt   = 0;

		ugcp->usbgc_tx_auto_pad   = B_FALSE;
		ugcp->usbgc_tx_list_max   = 16;

		ugcp->usbgc_rx_header_len = 4;
		ugcp->usbgc_rx_list_max   = 16;

		/* time out parameters */
		ugcp->usbgc_tx_timeout = 2*ONESEC;
		ugcp->usbgc_tx_timeout_interval = ONESEC;

		/* flow control */
		ugcp->usbgc_flow_control = FLOW_CONTROL_RX_PAUSE;

		/* MII timeout parameters */
		ugcp->usbgc_mii_link_watch_interval = ONESEC;
		ugcp->usbgc_mii_an_watch_interval = ONESEC/5;
		ugcp->usbgc_mii_reset_timeout = MII_RESET_TIMEOUT; /* 1 sec */
		ugcp->usbgc_mii_an_timeout = MII_AN_TIMEOUT;	/* 5 sec */
		ugcp->usbgc_mii_an_wait = (25*ONESEC)/10;
		ugcp->usbgc_mii_linkdown_timeout = MII_LINKDOWN_TIMEOUT;

		ugcp->usbgc_mii_an_delay = ONESEC/10;
		ugcp->usbgc_mii_linkdown_action = MII_ACTION_RSA;
		ugcp->usbgc_mii_linkdown_timeout_action = MII_ACTION_RESET;
		ugcp->usbgc_mii_dont_reset = B_FALSE;

		/* I/O methods */

		/* mac operation */
		ugcp->usbgc_attach_chip = &upf_attach_chip;
		ugcp->usbgc_reset_chip = &upf_reset_chip;
		ugcp->usbgc_init_chip = &upf_init_chip;
		ugcp->usbgc_start_chip = &upf_start_chip;
		ugcp->usbgc_stop_chip = &upf_stop_chip;
		ugcp->usbgc_multicast_hash = &upf_mcast_hash;

		ugcp->usbgc_set_rx_filter = &upf_set_rx_filter;
		ugcp->usbgc_set_media   = &upf_set_media;
		ugcp->usbgc_get_stats   = &upf_get_stats;
		ugcp->usbgc_interrupt   = NULL;

		/* packet operation */
		ugcp->usbgc_tx_make_packet = &upf_tx_make_packet;
		ugcp->usbgc_rx_make_packet = &upf_rx_make_packet;

		/* mii operations */
		ugcp->usbgc_mii_init   = &usbgem_mii_init_default;
		ugcp->usbgc_mii_config = &usbgem_mii_config_default;
		ugcp->usbgc_mii_sync  = &upf_mii_sync;
		ugcp->usbgc_mii_read  = &upf_mii_read;
		ugcp->usbgc_mii_write = &upf_mii_write;

		ugcp->usbgc_mii_tune_phy = NULL;

		lp = kmem_zalloc(sizeof (struct upf_dev), KM_SLEEP);

		ddi_set_driver_private(dip, NULL);

		dp = usbgem_do_attach(dip, ugcp, lp, sizeof (struct upf_dev));

		kmem_free(ugcp, sizeof (*ugcp));

		if (dp != NULL) {
			upf_ppa++;
			return (DDI_SUCCESS);
		}

err_free_mem:
		kmem_free(lp, sizeof (struct upf_dev));
err_close_pipe:
err:
		return (DDI_FAILURE);
	}
	return (DDI_FAILURE);
}

static int
upfdetach(dev_info_t *dip, ddi_detach_cmd_t cmd)
{
	int	ret;

	if (cmd == DDI_DETACH) {
		ret = usbgem_do_detach(dip);
		if (ret != DDI_SUCCESS) {
			DDI_FAILURE;
		}
		upf_ppa--;
		return (DDI_SUCCESS);
	}
	return (DDI_FAILURE);
}

/* ======================================================== */
/*
 * OS depend (loadable streams driver) routine
 */
/* ======================================================== */
static	struct module_info upfminfo = {
	0,			/* mi_idnum */
	"upf",			/* mi_idname */
	0,			/* mi_minpsz */
	ETHERMTU,		/* mi_maxpsz */
	32*1024,		/* mi_hiwat */
	1,			/* mi_lowat */
};

static	struct qinit upfrinit = {
	(int (*)()) NULL,	/* qi_putp */
	usbgem_rsrv,		/* qi_srvp */
	usbgem_open,		/* qi_qopen */
	usbgem_close,		/* qi_qclose */
	(int (*)()) NULL,	/* qi_qadmin */
	&upfminfo,		/* qi_minfo */
	NULL			/* qi_mstat */
};

static	struct qinit upfwinit = {
	usbgem_wput,		/* qi_putp */
	usbgem_wsrv,		/* qi_srvp */
	(int (*)()) NULL,	/* qi_qopen */
	(int (*)()) NULL,	/* qi_qclose */
	(int (*)()) NULL,	/* qi_qadmin */
	&upfminfo,		/* qi_minfo */
	NULL			/* qi_mstat */
};

static struct streamtab	upf_info = {
	&upfrinit,	/* st_rdinit */
	&upfwinit,	/* st_wrinit */
	NULL,		/* st_muxrinit */
	NULL		/* st_muxwrinit */
};

static	struct cb_ops cb_upf_ops = {
	nulldev,	/* cb_open */
	nulldev,	/* cb_close */
	nodev,		/* cb_strategy */
	nodev,		/* cb_print */
	nodev,		/* cb_dump */
	nodev,		/* cb_read */
	nodev,		/* cb_write */
	nodev,		/* cb_ioctl */
	nodev,		/* cb_devmap */
	nodev,		/* cb_mmap */
	nodev,		/* cb_segmap */
	nochpoll,	/* cb_chpoll */
	ddi_prop_op,	/* cb_prop_op */
	&upf_info,	/* cb_stream */
	D_MP		/* cb_flag */
};

static	struct dev_ops upf_ops = {
	DEVO_REV,	/* devo_rev */
	0,		/* devo_refcnt */
	usbgem_getinfo,	/* devo_getinfo */
	nulldev,	/* devo_identify */
	nulldev,	/* devo_probe */
	upfattach,	/* devo_attach */
	upfdetach,	/* devo_detach */
	nodev,		/* devo_reset */
	&cb_upf_ops,	/* devo_cb_ops */
	NULL,		/* devo_bus_ops */
	ddi_power	/* devo_power */
};

static struct modldrv modldrv = {
	&mod_driverops,	/* Type of module.  This one is a driver */
	ident,
	&upf_ops,	/* driver ops */
};

static struct modlinkage modlinkage = {
	MODREV_1, &modldrv, NULL
};

/* ======================================================== */
/*
 * _init : done
 */
/* ======================================================== */
int
_init(void)
{
	int 	status;

	DPRINTF(2, (CE_CONT, "!upf: _init: called"));
	status = mod_install(&modlinkage);

	return (status);
}

/*
 * _fini : done
 */
int
_fini(void)
{
	int	status;

	DPRINTF(2, (CE_CONT, "!upf: _fini: called"));
	status = mod_remove(&modlinkage);
	return (status);
}

int
_info(struct modinfo *modinfop)
{
	return (mod_info(&modlinkage, modinfop));
}
